#include "rpi_cava.h"

#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <pthread.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#define MAXLINECHAR 128 * (5 + 1) + 1

pid_t cavaPid;
bool isFifoReaderRunning = false;
pthread_t fifoReader;
int cavaFifo;
char lineBuffer[MAXLINECHAR];
uint16_t buffer[128 + 2];

static void *fifo_to_buffer(void *arg);

void setup_cava() {
  if ((cavaPid = fork()) == -1) {
    perror("fork");
    exit(1);
  }

  if (cavaPid == 0) {
    /* Child process*/
    pthread_setschedprio(pthread_self(), 30);
    char *args[] = {"/usr/local/bin/cava", "-p", "/home/pi/LedBars/RpiLedBars/cava_config", NULL};

    if (execv(args[0], args) != 0) {
      perror("execv");
    }

  } else {
    sleep(1);
  }
}

void close_cava() {
  stop_cava_bg_worker();
  kill(cavaPid, SIGTERM);
}

void start_cava_bg_worker() {
  isFifoReaderRunning = true;
  pthread_create(&fifoReader, NULL, fifo_to_buffer, NULL);
}

void stop_cava_bg_worker() {
  isFifoReaderRunning = false;
  if (pthread_join(fifoReader, NULL) != 0) {
    perror("pthread_join");
  }
}

int get_cava_buffer(uint16_t **buffer_dst) {
  *buffer_dst = buffer;
  return 0;
}

size_t valueIndex = 0, charOffset = 0;
char strValue[6] = "0\0";
bool hasToBeDiscarded = false;

static void *fifo_to_buffer(void *arg) {
  struct pollfd fds[1];
  int timeoutMs;
  int ret;

  if ((cavaFifo = open("/tmp/cava_output", O_RDONLY | O_NONBLOCK)) < 0) {
    perror("open");
    close_cava();
    exit(1);
  }

  memset(fds, 0, sizeof(fds));
  fds[0].fd = cavaFifo;
  fds[0].events = POLLIN;
  timeoutMs = 10;

  hasToBeDiscarded = true;

  while (isFifoReaderRunning) {

    if ((ret = poll(fds, 1, timeoutMs)) == 1) {
      int nread;
      nread = read(cavaFifo, lineBuffer, 128 + 1);

      if (nread >= 0) {
        for (size_t i = 0; i < nread; ++i) {
          char current = lineBuffer[i];

          if (hasToBeDiscarded) {
            if (current == '\n') {
              charOffset = 0;
              strValue[charOffset] = '\0';
              valueIndex = 0;
              hasToBeDiscarded = false;
            }

          } else {
            if ('0' <= current && current <= '9') {
              strValue[charOffset++] = current;
            } else if (current == '\n' || current == ';') {
              strValue[charOffset] = '\0';
              charOffset = 0;
              buffer[valueIndex++] = atoi(strValue);

              if (current == '\n' || valueIndex > 129) {
                valueIndex = 0;

                if (valueIndex > 129) {
                  fprintf(stderr, "Buffer overflow, \\n missed, discarding next\n");
                  hasToBeDiscarded = true;
                }
              }

            } else {
              fprintf(stderr, "Unexpected char %d [%c]\n", current, current);
            }
          }
        }
      } else {
        if (errno != EAGAIN) {
          perror("read");
        }
      }
    } else if (ret < 0) {
      fprintf(stderr, "error polling %d: %s\n", cavaFifo, strerror(errno));
    }
  }

  close(cavaFifo);

  return NULL;
}
